import serial
import time
import argparse
import fsm
import gen_header

PEER_ID = 2
BAUDRATE = 115200

# ================================================================================================
def time_after(a, b):
  return int(b) - int(a) < 0

#####################################################
OTA_MAGIC = 0x534E4150
RET_OK = 0x00
RET_ERROR = 0x01
CMD_GET_VER = 0x11
CMD_OTA_CMD_SET = 0xAD
CMD_OTA_START = 0x01
CMD_OTA_TRNAS = 0x02
CMD_OTA_FINISH = 0x03

CMD_SET_SYS = 0x01
CMD_ID_SYS_SET_PC_PORT_PRO = 0x11

def get_seq():
  global seq
  ret = seq
  seq += 1
  seq = seq & 0xff
  return ret

#####################################################
def clear_ser(ser):
  while True:
    c = ser.read()
    if 0 == len(c):
      return
    else:
      print(c)

def get_frame(ser, sf):
  start_tick_ms = int(round(time.time() * 1000))
  while True:
    c = ser.read()
    if len(c):
      if sf.push_char(c[0]):
        frame_str = " ".join(["{:02x}".format(x) for x in sf.frame])
        print("RX : " + frame_str)
        return sf.frame
    cur = int(round(time.time() * 1000))
    if time_after(cur, start_tick_ms + 4000):
      print("get_frame timeout!")
      return None

def _4_byte_xor(data):
  l = int(len(data)>>2)
  if len(data)%4:
    cutoff = data[l*4:]
    while True:
      if len(cutoff) < 4:
        cutoff.extend(0)
      else:
        break
  else:
    cutoff = None

  ret = bytearray(4)
  for i in range(l):
    ret[0] ^= data[i*4+0]
    ret[1] ^= data[i*4+1]
    ret[2] ^= data[i*4+2]
    ret[3] ^= data[i*4+3]
  if cutoff:
    ret[0] ^= cutoff[0]
    ret[1] ^= cutoff[1]
    ret[2] ^= cutoff[2]
    ret[3] ^= cutoff[3]
  return ret

def reset_target(ser, sf):
  pl = bytearray()
  pl.append(0x01)
  pl.append(0x03)
  frame = sf.build_frame(pl, PEER_ID)
  frame_str = " ".join(["{:02x}".format(x) for x in frame])
  print("TX: " + frame_str)
  ser.write(frame)

def ota_start(ser, bin_data, sf):
  pl = bytearray()
  pl.append(CMD_OTA_CMD_SET)
  pl.append(CMD_OTA_START)
  pl.append(0x00)
  pl.append(0x01)
  pl.extend(bin_data[0:256])
  frame = sf.build_frame(pl, PEER_ID)
  frame_str = " ".join(["{:02x}".format(x) for x in frame])
  print("TX: " + frame_str)
  ser.write(frame)

  ret_frame = get_frame(ser, sf)
  if None == ret_frame:
    return False

  cmd_set, cmd_id, ret_code = sf.pars_ret_frame(ret_frame)
  if cmd_set != CMD_OTA_CMD_SET:
    print("ota start ack code error")
    return False

  if cmd_id != CMD_OTA_START:
    print("ota start ack code error")
    return False

  if ret_code != RET_OK:
    print("ota start return %d" % ret_code)
    return False

  return True

def ota_trans(ser, offset, block, sf):
  pl = bytearray()
  pl.append(CMD_OTA_CMD_SET)
  pl.append(CMD_OTA_TRNAS)
  pl.append(0)
  pl.extend(offset.to_bytes(4, 'little'))
  pl.extend(len(block).to_bytes(2, 'little'))
  pl.extend(block)
  frame = sf.build_frame(pl, PEER_ID, 1)
  # frame_str = " ".join(["{:02x}".format(x) for x in frame])
  # print(frame_str)
  ser.write(frame)
  return None

def ota_finish(ser, ret, sf):
  pl = bytearray()
  pl.append(CMD_OTA_CMD_SET)
  pl.append(CMD_OTA_FINISH)
  pl.append(ret & 0xFF)
  frame = sf.build_frame(pl, PEER_ID, 1)
  # frame_str = " ".join(["{:02x}".format(x) for x in frame])
  # print(frame_str)
  ser.write(frame)
  return None

def pc_port_change_to_sacp(ser, sf):
  pl = bytearray()
  pl.append(CMD_SET_SYS)
  pl.append(CMD_ID_SYS_SET_PC_PORT_PRO)
  pl.append(0x00)
  pl.append(0x01)
  pl.append(0x01)
  frame = sf.build_frame(pl, PEER_ID)
  frame_str = " ".join(["{:02x}".format(x) for x in frame])
  print("TX: " + frame_str)
  ser.write(frame)

  ret_frame = get_frame(ser, sf)
  if None == ret_frame:
    return False

  cmd_set, cmd_id, ret_code = sf.pars_ret_frame(ret_frame)
  if cmd_set != CMD_SET_SYS:
    return False

  if cmd_id != CMD_ID_SYS_SET_PC_PORT_PRO:
    return False

  if ret_code != RET_OK:
    return False

  return True

def ota(ser, bin_data, sf):
  # if True != pc_port_change_to_sacp(ser, sf):
  #   print("can not set pc port to SACP protocol")
  #   return False
  pl = "M1999\n".encode('utf-8')
  print(pl)
  ser.write(pl)
  ser.write(pl)
  ser.write(pl)
  bin_size = len(bin_data) - 256
  start_time = time.time()
  for i in range(3):
    if True == ota_start(ser, bin_data, sf):
      break

  if i == 2:
    return False
  time_out_cnt = 0
  while True:
    ret_code = RET_ERROR
    ret_frame = get_frame(ser, sf)
    if None == ret_frame:
      time_out_cnt += 1
      if time_out_cnt > 20:
        break
      continue

    cmd_set, cmd_id, ret_code = sf.pars_ret_frame(ret_frame)
    if cmd_set != CMD_OTA_CMD_SET:
      continue

    if cmd_id == CMD_OTA_TRNAS:
      req_offset = int.from_bytes(ret_frame[13:17], 'little')
      block_size = int.from_bytes(ret_frame[17:19], 'little')
      print("req offset %d, block size %d, progress %f%%, time escape %d second" % (req_offset, block_size, 100 * req_offset/bin_size, time.time() - start_time))
      block = bin_data[256 + req_offset : 256 + req_offset + block_size]
      ota_trans(ser, req_offset, block, sf)

    elif cmd_id == CMD_OTA_FINISH:
      ota_finish(ser, ret_code, sf)
      break

  if ret_code == RET_OK:
    return True
  else:
    return False

def main_run():
  parser = argparse.ArgumentParser(description="sacp upgrade")
  parser.add_argument('--com', '-p', help='specify the serial port', default='COM1')
  parser.add_argument('--baud', '-b', help='baudrate', default=115200)
  parser.add_argument('--file', '-f', help='file', default='./firmware.bin')
  args = parser.parse_args()
  try:
    com = args.com
    baudrate = int(args.baud)
    raw_firmware_file = args.file
  except:
    print("unsupported parameters! Please check the help information by run 'python sacp_upgrade.py -h'")
    while True:
      pass

  # Generate the packet for upgrade by adding Header
  with open(raw_firmware_file, 'rb') as f:
    raw_bin = f.read()
  hg = gen_header.packet(raw_bin, 3, 0, 'V1.1.0', 0x8002800, 0, 0)
  packet_bin = hg.gen()
  packet_bin.extend(raw_bin)

  # ================================================================================================
  ser = serial.Serial(com, baudrate, timeout=0.5)
  sf = fsm.fsm()
  seq = 0
  bin_file = args.file

  clear_ser(ser)
  if ota(ser, packet_bin, sf):
    print("OK: app upgrade application seccussful!")
  else:
    print("ERR: app upgrade application failed!")

if __name__=='__main__':
  main_run()